/* pppox.c - pppd plugin to implement PPPoX protocol.
 *
 * By Martijn van Oosterhout <kleptog@svana.org>
 *
 *
 * Heavily based upon pppoatm.c: original notice follows
 *
 * Copyright 2000 Mitchell Blank Jr.
 * Based in part on work from Jens Axboe and Paul Mackerras.
 * Updated to ppp-2.4.1 by Bernhard Kaindl
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "pppd.h"
#include "pathnames.h"
#include "fsm.h" /* Needed for lcp.h to include cleanly */
#include "lcp.h"
#include <sys/stat.h>
#include <net/if.h>
#include <sys/ioctl.h>

const char pppd_version[] = VERSION;

static int setdevname_pppox(const char **argv);

//static int device_got_set = 0;
static int pppox_max_mtu = 1460, pppox_max_mru = 1460;
static int pppox_fd = -1;
static char *pppox_fd_str;
static int device_got_set = 0;
struct channel pppox_channel;

static option_t pppox_options[] = {
	{ "pppox", o_special, &setdevname_pppox,
	  "FD for PPPoX socket", OPT_DEVNAM | OPT_A2STRVAL,
          &pppox_fd_str },
	{ NULL }
};

static void no_device_given_pppoatm(void)
{
	fatal("No vpi.vci specified");
}

static int setdevname_pppox(const char **argv)
{
	union {
		char buffer[128];
		struct sockaddr pppox;
	} s;
	int len = sizeof(s);

	info( "setdevname_pppox(%p)\n", argv );
	info( "setdevname_pppox(%s)\n", argv[0] );
		
	if (device_got_set)
		return 0;
		
	if (!int_option(*argv, &pppox_fd))
		return 0;
	                
	if( getsockname( pppox_fd, (struct sockaddr *)&s, &len ) < 0 )
	{
		fatal( "Given FD for PPPoX socket invalid (%s)", strerror(errno) );
	}
	if( s.pppox.sa_family != AF_PPPOX )
	{
		fatal( "Socket of not a PPPoX socket" );
	}
	
	the_channel = &pppox_channel;
	device_got_set = 1;
	info( "PPPoX using FD %d", pppox_fd );
	return 1;
}

static int connect_pppox(void)
{
	if( pppox_fd == -1 )
	{
		fatal( "No PPPoX FD specified" );
	}
	return pppox_fd;
	
}

static void disconnect_pppox(void)
{
	close(pppox_fd);
	pppox_fd = -1;
}

static void send_config_pppox(int mtu,
			      u_int32_t asyncmap,
			      int pcomp,
			      int accomp)
{
	struct ifreq ifr;
	if (mtu > pppox_max_mtu)
		error("Couldn't increase MTU to %d (exceeds max %d)", mtu, pppox_max_mtu);
	if (ioctl(pppox_fd, SIOCSIFMTU, (caddr_t) &ifr) < 0)
		fatal("ioctl(SIOCSIFMTU): %m");
}

static void recv_config_pppox(int mru,
			      u_int32_t asyncmap,
			      int pcomp,
			      int accomp)
{
	if (mru > pppox_max_mru)
		error("Couldn't increase MRU to %d", mru);
}

void plugin_init(void)
{
#if defined(__linux__)
	extern int new_style_driver;	/* From sys-linux.c */
	if (!ppp_available() && !new_style_driver)
		fatal("Kernel doesn't support ppp_generic - "
		    "needed for PPPoATM");
#else
	fatal("No PPPoX support on this OS");
#endif
	info("PPPoX plugin_init");
	add_options(pppox_options);
}
struct channel pppox_channel = {
    options: pppox_options,
    process_extra_options: NULL,
    check_options: NULL,
    connect: &connect_pppox,
    disconnect: &disconnect_pppox,
    establish_ppp: &generic_establish_ppp,
    disestablish_ppp: &generic_disestablish_ppp,
    send_config: &send_config_pppox,
    recv_config: &recv_config_pppox,
    close: NULL,
    cleanup: NULL
};
